require "narray_yaml"
require "numru/gphys"
require "numru/netcdf"

class Variable < ActiveRecord::Base
  belongs_to :directory
  belongs_to :user
  belongs_to :group
  has_many :keyword_attributes, :dependent => :destroy
  has_many :spatial_attributes, :dependent => :destroy
  has_many :references, :through => :variable_relations_reference
  has_many :cited_by, :through => :variable_relations_cited_by
  has_many :variable_relations_reference, :class_name => "VariableRelation", :foreign_key => :cited_by
  has_many :variable_relations_cited_by, :class_name => "VariableRelation", :foreign_key => :reference
  has_many :diagram_caches, :through => :diagram_cache_data, :source => :diagram_cache
  has_many :diagram_cache_data, :dependent => :destroy
  has_many :actual_files, :dependent => :destroy
  has_many :draw_parameters, :dependent => :destroy
  has_many :knowledge_figures

  DIR_VAR_DELIMITER = '@'

  validates_presence_of  :path, :name
  validates_uniqueness_of :path
  validates_associated :directory
  def before_create
    make_directories unless directory
  end

  def after_save
    if @references_tmp
      @references_tmp.each{|v|
#        references << v
        VariableRelation.new(:reference => v, :cited_by => self).save!
      }
    end
    yaml_fname = file_and_variable_names[0] + ".yml"
    unless File.exist?(yaml_fname)
      if keyword_attributes.length > 0 || user_id || references.length > 0
        hash = Hash.new
        keyword_attributes.each{|key| hash[key.name] = key.value}
        if user_id && user_id > 0
          hash["gfdnavi"] ||= Hash.new
          hash["gfdnavi"]["user"] = User.find(user_id).login
          if group_id && (group = Group.find(group_id)) && group.user_id == user_id
            hash["gfdnavi"]["group"] = group.name
          end
        end
        if references.length > 0
          hash["gfdnavi"]["references"] = references.collect{|ref|
            {
              "path" => ref.path
            }
          }
        end
        hash = {"contains" => {self.name => hash}}
        File.open(yaml_fname,"w"){|file| file.print(hash.to_yaml) }
      end
    end
  end


#  def self.find(*args)
#    raise NoMethodError, "This method is diabled. Use find_for_user"
#  end

  def self.find_for_user(user, *args)
    args = [:all] if args.length == 0
    if user && user.login == "root"
      return self.find(*args)
    end

    grp_cond = Group.group_conditions(user)

    if args.length == 1
      args.push({:conditions => grp_cond})
    elsif args[1].is_a?(Hash)
      if (cond = args[1][:conditions])
        if Array === cond
          args[1][:conditions] = ( ['( '+ cond[0] +' ) AND ' + grp_cond ] + cond[1..-1] )
        else
          args[1][:conditions] = '( '+ cond +' ) AND ' + grp_cond
        end
      else
        args[1][:conditions] = grp_cond
      end
    else
      # args must be IDs --> put them in an array
      args.each do |x|
        if !x.is_a?(Integer)
          raise ArgumentError,
            "ID must be integers, but #{x.inspect} isn't. Or you need to give"+
            "a Hash as the 2nd argument, which is now #{args[1].inspect}."
        end
      end
      args = [ args, {:conditions =>grp_cond} ]
    end

    return Variable.find(*args)

  end

  def name=(str)
    ##  gphys.name gƁAopen ςȂɂȂ̂ł͔B
    ##  Ƃ肠RgAEgB
    ##if gphys? && ( (nm=gphys.name) == str )
    ##  raise ArgumentError, "Name miss-match. '#{str}' != '#{nm}' (in file)"
    ##end
    super(str)
  end

  def file_and_variable_names
    if gphys?
      /^(.*)#{DIR_VAR_DELIMITER}(.*?)$/ =~ path
      fn = $1
      vname = $2
      afs = actual_files
      if afs.length > 0
	fnames = afs.collect{|af| af.path}
      else
	fnames = [ fn ]
      end
      raise "Invalid path (#{path}). Could't separate file&var names" if !vname
    else
      fnames = [ path ]
      vname = nil
    end

    fnames.each_with_index do |fn,idx|
      if /^temporary:(.*)/ =~ fn
	path = $1
	if vname
	  fnames[idx] = GFDNAVI_WORK_PATH + path
	else
	  fnames[idx] = GFDNAVI_DIAGRAM_PATH + path
	end
      elsif /^\/usr\/([^\/]*)\/(.*)/ =~ fn
	uname = $1
	dir = $2
	fnames[idx] = GFDNAVI_USER_PATH + "/" + uname + "/" + dir
      else
	fnames[idx] = GFDNAVI_DATA_PATH + fn
      end
    end

    if fnames.length > 1
      return [fnames, vname]
    elsif fnames.length == 1
      return [fnames.first, vname]
    else
      raise "No file found. Sorry, this is likely because of a bug."
    end
  end

  def gphys?
    return gphysizable && path!=""
  end

  def gphys
    unless gphys?
      raise "This is not GPhysizeable data"
    end
    fname, vname = file_and_variable_names()
    obj = NumRu::GPhys::IO.open(fname, vname)
  end

  def references_tmp
    unless @new_record
      raise "Cannot #call references_tmp for saved data. Use references instead"
    end
    @references_tmp ||= Array.new
  end

  protected
  def validate
    if path =~ /^temporary/
      errors.add(:path, "temporary variable cannot be saved")
    end
    tmp = path.split(":")
    if tmp.length == 1
      tmp = tmp[0]
    elsif tmp.length == 2
      tmp = tmp[1]
    else
      errors.add(:path, "format of path is imvalid")
    end
    unless tmp =~ /^\//
      errors.add(:path, "path must be started with '/'")
    end
  end

  def make_directories
    dir = Directory.find_by_name("/")
    apath = path.split(DIR_VAR_DELIMITER)[0]
    if /^\/usr\/([^\/]*)\/(.*)/ =~ apath
      user = $1
      apath = $2
      dir = make_directory(dir, "usr")
      dir = make_directory(dir, user)
    else
      apath = apath[1..-1]
    end
    apath = apath.split("/")
    apath.each{|name|
      dir = make_directory(dir, name)
    }
    self.directory = dir
  end

  def make_directory(parent, name)
    children = parent.children
    parent_path = parent.path
    parent_path += "/" unless parent_path =~ /\/$/
    return children.find_by_name(name) || children.create(:name => name, :path => parent_path+name)
  end

end
