#!/usr/bin/env ruby
require "date"
class DateTime
  alias __to_s to_s
  def to_s
    __to_s.sub(/Z$/,'')
  end
end

begin # for RubyGems users
  require "rubygems"
  gem "activerecord"
rescue LoadError
  # for Debian package users
  $LOAD_PATH << "/usr/share/rails/activerecord/lib/"
  require "active_record"
end
require "kconv"
require "find"
require "optparse"
require "yaml"
require "numru/gphys"

class OptionParser
  # Changed to raise error. (Originally, invalid options
  # are just warned, which is easily overlooked.)
  def warn(mesg = $!)  
    raise mesg
  end
end

CONFIG_PATH = File.join(File.dirname(__FILE__), '../config')
require File.join(CONFIG_PATH, "environment")

app_path = File.join(File.dirname(__FILE__), '../app')
require File.join(app_path, "models/directory")
require File.join(app_path, "models/keyword_attribute")
require File.join(app_path, "models/variable")
require File.join(app_path, "models/spatial_attribute")
require File.join(app_path, "models/actual_file")
require File.join(app_path, "models/draw_parameter")
require File.join(app_path, "models/variable_relation")
require File.join(app_path, "models/user")
require File.join(app_path, "models/group")

CONFIG_DATABASE = YAML.load_file(File.join(CONFIG_PATH, "database.yml"))
DEFAULT_DATABASE = 'development'

GFDNAVI_CHARSET = Kconv::UTF8
GFDNAVI_LANGUAGES = ['ja', 'en']


# o^XNvg Web C^[tF[XƂŋL̃L[
YAML_GFDNAVI_SHARED_KEYWORD = 'gfdnavi'

# f[^t@Co^p^t@CȊO̓o^t@C̊gq
NON_GPHYSIZABLE_SUFFIXES = /\.png$/


module RegisterSharedTools
  #
  #hash ɓo^L[[hƒl̃Zbg^, dir_id, var_id ɂ
  #KeywordAttribute  belong_to ̊֌WɂfBNg܂
  #ϐ ID ^.
  #
  #L[[h "gfdnavi" ^ꍇ, ̒l gfdnavi  
  #Web C^[tF[Xɂč쐬ꂽƂēo^܂.
  #(̓Iɂ gfdnavi_params_register ɓn, œo^
  #s܂).
  #
  #L[[hƒl 2 oCg܂܂ꍇ, GFDNAVI_CHARSET
  #Őݒ肳镶R[hɕϊēo^.
  #
  # lang ݒ肳Ăꍇ,  lang ɐݒ肳镶
  #GFDNAVI_LANGUAGES ɓo^Ă̂Ȃ, ̕
  #keyword_attribute e[u lang ɓo^.
  #
  #lang  nil łꍇɂ, L[[h̖ "-ja" ̂悤
  #񂪂ꍇ, ̌ݒ肪 keyword_attribute e[u
  #lang ɓo^. "-ja" ̃TtBbNX͎菜.
  #
  #lang  nil ł, L[[h̖ "-ja" ̂悤Ȍ
  #TtBbNXȂꍇł, l 2 oCgłꍇɂ
  #"ja"  keyword_attribute e[u lang ɓo^.
  #
  #߂l, Vɓo^ꂽ KeywordAttribute  ID ̃Xg
  #ł. VKɓo^ꂽ̂ꍇ͋ Array Ԃ.
  #
  def keyword_attribute_register(hash, dir_id=nil, var_id=nil, lang=nil)
    keyatt_ids = []
    return keyatt_ids if !(dir_id) && !(var_id)
    hash.each{ |key, value|

      #-------------------------------------------
      #  do^̉
      #-------------------------------------------

      if dir_id
	recorded_keyatt = KeywordAttribute.find(:first, :conditions => \
	        ['(directory_id = ?) and (name = ?)',dir_id.to_s, key])
      else
	recorded_keyatt = KeywordAttribute.find(:first, :conditions => \
	        ['(variable_id = ?) and (name = ?)',var_id.to_s, key])
      end
      next if recorded_keyatt

      #-------------------------------------------
      #  o^
      #-------------------------------------------

      if key == YAML_GFDNAVI_SHARED_KEYWORD
        keyatt_ids << gfdnavi_params_register(value, dir_id, var_id)
        keyatt_ids.delete_if {|k| !(k)}
        keyatt_ids.flatten!
      else
        if value.class.to_s == "Hash"
          keyword_attribute_register(value, dir_id, var_id, lang)
        end
        keyatt = KeywordAttribute.new
        keyatt.directory_id = dir_id if dir_id
        keyatt.variable_id = var_id if var_id
        keyatt.lang = lang if lang && GFDNAVI_LANGUAGES.include?(lang)
        suffix = nil
        if !(keyatt.lang) && key =~ /\-(\w+)$/
          suffix = $1
          if GFDNAVI_LANGUAGES.include?(suffix)
            keyatt.lang = suffix
            key = key.sub(/\-\w+$/, '')
          end
        end
        if key.class.to_s == "String"
          keyatt.name = Kconv.kconv(key, GFDNAVI_CHARSET)
          if !(keyatt.lang) && ja_charset?(key)
            keyatt.lang = 'ja'
          end
        else
          keyatt.name = key
        end
        if value.class.to_s == "String"
          keyatt.value = Kconv.kconv(value, GFDNAVI_CHARSET)
          if !(keyatt.lang) && ja_charset?(value)
            keyatt.lang = 'ja'
          end
        elsif value.class.to_s == "NilClass"
          keyatt.value = ''
        elsif value.class.to_s == "NArray"
          keyatt.value = value
        else
          begin
            keyatt.value = NArray.to_na(value)
          rescue
            keyatt.value = value.to_s
          end
        end
        keyatt.save
        keyatt_ids << keyatt.id
      end
    }
    return keyatt_ids
  end

  #
  #str ɗ^镶񂪓{ł true Ԃ.
  #
  def ja_charset?(str)
    guessed_code = Kconv.guess(str)
    if [Kconv::JIS, Kconv::EUC, Kconv::SJIS,
        Kconv::UTF8, Kconv::UTF16].include?(guessed_code)
      return true
    else
      return false
    end
  end

  #
  #hash ɓo^L[[hƒl̃Zbg^, dir_id, var_id ɂ
  #KeywordAttribute  belong_to ̊֌WɂfBNg܂
  #ϐ ID ^.
  #
  #hash  HASH ȊÕNX^ꍇ, G[Ԃ.
  #
  def gfdnavi_params_register(hash, dir_id=nil, var_id=nil)
    return nil unless hash
    return nil if !(dir_id) && !(var_id)
    entity_id = dir_id || var_id
    keyatt_ids = []

    #-------------------------------------------
    #  YAML t@C gfdnavi ̒l HASH ŖꍇɃG[
    #-------------------------------------------

    if dir_id
      dir_ob = Directory.find(:first, :conditions=>["id = ?", dir_id]) 
    elsif
      dir_ob = Variable.find(:first, :conditions=>["id = ?", var_id]).directory
    end
    entity_filename = dir_ob.path
    yml_filename = entity_filename + '.yml'

    unless hash.class.to_s == "Hash"
      raise GfdnaviYamlSyntaxError, "  Error: \"#{YAML_GFDNAVI_SHARED_KEYWORD}\" information in \"#{yml_filename}\" must be Hash format.\n"
    end

    #-------------------------------------------
    #  o^
    #-------------------------------------------

    hash.each{ |key, value|
      if key == 'draw_parameters'
        unless value.class.to_s == "Hash"
          raise GfdnaviYamlSyntaxError, "  Error: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: draw_parameters\" information in \"#{yml_filename}\" must be HashH format.\n"
        end
        if !(var_id)
          warn "  Warning: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: draw_parameters\" information in \"#{yml_filename}\" can " +
            "not be added to Directory\n"
          next
        end

        value.each{|vkey, vvalue|
          next if DrawParameter.find( :first, :conditions=> \
                   ["(variable_id = ?) and (name = ?)", entity_id.to_s, vkey] )

          dp = DrawParameter.new
          dp.name = vkey
          dp.value = vvalue
          dp.variable_id = entity_id
          dp.save!
          keyatt_ids << dp.id
        }
      elsif key == 'user' || key == 'group'
        unless value.is_a?(String)
          raise GfdnaviYamlSyntaxError, "  Error: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: #{key}\" information in \"#{yml_filename}\" must be String format.\n"
        end
        var = Variable.find(:first, :conditions=>["(id = ?)", entity_id.to_s])
        next if !var
        user_ob = nil
        group_ob = nil
        if key == 'user'
          user_ob = User.find(:first, :conditions=>["(login = ?)", value])
        else
          group_ob = Group.find(:first, :conditions=>["(name = ?)", value])
        end
        if user_ob
          var.user_id = user_ob.id
          var.save
        elsif group_ob
          var.group_id = group_ob.id
          var.save
        else
          warn "  Warning: #{key} \"#{value}\" in \"#{yml_filename}\" was lost from the database. Make it, and do directory-tree registration again.\n"
	  if key == 'group' 
	    gm = Group.find(:first, :conditions=>["name = 'group_missing'"])
	    var.group_id = gm.id if gm
	    var.save
	    warn "  --> #{var.path} will be available only to the super user"
	  end
        end
      elsif key == 'references'
        unless value.class.to_s == "Array"
          raise GfdnaviYamlSyntaxError, "  Error: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: #{key}\" information in \"#{yml_filename}\" must be Array format.\n"
        end
        if !(var_id)
          warn "  Warning: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: references\" information in \"#{yml_filename}\" can " +
            "not be added to Directory\n"
          next
        end
        value.each{|v|
          next if !(v.class.to_s == "Hash" && v.has_key?('path'))
          ref = 
            Variable.find(:first, :conditions=>["(path = ?)", v['path']])
          cited = 
            Variable.find:first, :conditions=>(["(id = ?)", var_id])

          if ref && cited
            vr = VariableRelation.new
            vr.reference = ref
            vr.cited_by = cited
            vr.save
          end
        }
      else
        warn "  Warning: \"#{YAML_GFDNAVI_SHARED_KEYWORD}: #{key}\" information in \"#{yml_filename}\" is ignored.\n"
      end
    }
    keyatt_ids
  end

  class GfdnaviYamlSyntaxError < Exception
  end

  module_function :keyword_attribute_register
  module_function :gfdnavi_params_register
  module_function :ja_charset?
end

#
#fBNgȉ̃fBNg񂨂т̃fBNgɕt
#SIGEN t@C gfdnavi f[^x[Xɓo^邽߂̃NX.
#
class RegisterDirTree

  # [gfBNg̖
  ROOTDIR_NAME = '/'
  ROOTPATH_NAME = '/'

  # fBNgƕϐ̋؂蕶
  DIR_VAR_DELIMITER = Variable::DIR_VAR_DELIMITER

  # YAML t@C, GPhys ϐɏt邽߂̃L[
  YAML_GPHYSVAR_KEYWORD = 'contains'

  #
  #\bh.
  #
  #topdir ɂ̓gbvfBNg^. ftHg,
  #config/gfdnavi.yml  data_path ɐݒ肳lł.
  #
  #userdir ɂ̓[U񂪕ۑfBNg̃gbvfBNg
  #^. ftHg config/gfdnavi.yml  user_path ɗ^
  #lł.
  #
  #database ɂ 'development', 'production', 'test' ̂ꂩ
  #^. ftHg 'development' ł.
  #
  #quiet  true ^, G[ȊÕbZ[Wo͂sȂ.
  #verbose  true ^, i璷ɏo͂.
  #ftHg͊ȒPȐi󋵂̏o͂s.
  #
  #clean_dir ɂ, fBNgw肷. w肳ꂽfBNg
  #͓o^OɈx폜. 폜ɔ, ̃fBNgȉ
  #o^e͈USč폜. 
  #ftHgł͊ɓo^Ă͍ēo^Ȃ, 
  #f[^ YAML t@C̒ĝݍXVꂽꍇ, 
  #̍XVef[^x[XɔfȂ. 
  #̂悤ȏꍇɂ clean_dir ɂt@C̐efBNg
  #w肷Ɨǂ. ̃fBNgw肷ۂ, 
  #fBNgJ} ',' ŋ؂.
  #
  #search_dir ɂ, o^̂߂̒TJnfBNgw肷. 
  #ftHg topdir Ŏw肳ꂽfBNgł. 
  #o^̂߂̃fBNgT̎ԒZks߂Ɏgp.
  #
  def initialize(topdir=GFDNAVI_DATA_PATH, userdir=GFDNAVI_USER_PATH, 
                 database=nil, quiet=nil, verbose=nil, 
                 clean_dir=nil, search_dir=nil)
    topdir ||= GFDNAVI_DATA_PATH
    raise IOError, "  Error: \"#{topdir}\" is not found.\n" if !(File.exist?(topdir))
    raise IOError, "  Error: \"#{topdir}\" is not a directory.\n" if !(File.directory?(topdir))
    @topdir = File.expand_path(topdir)
    @topdir.sub!(/\/$/, '')

    userdir ||= GFDNAVI_USER_PATH
    raise IOError, "  Error: \"#{userdir}\" is not found.\n" if !(File.exist?(userdir))
    raise IOError, "  Error: \"#{userdir}\" is not a directory.\n" if !(File.directory?(userdir))
    @userdir = File.expand_path(userdir)
    @userdir.sub!(/\/$/, '')
    @userdir_already_search = false

    @verbose_level = 1  # 0:quiet, 1: normal, 2: verbose
    @verbose_level = 0 if quiet
    @verbose_level = 2 if verbose
    normal_msg "\"#{topdir}\" is set as top directory.\n"

    @database = database || ENV['RAILS_ENV'] || DEFAULT_DATABASE

    ActiveRecord::Base.establish_connection(
                       :adapter => CONFIG_DATABASE[@database]['adapter'],
                       :host => CONFIG_DATABASE[@database]['host'],
                       :username => CONFIG_DATABASE[@database]['username'],
                       :password => CONFIG_DATABASE[@database]['password'],
                       :database => CONFIG_DATABASE[@database]['database']
                                            )

    #---------------------------------------------
    #  o^e̍폜 (clean_dir ^Ăꍇ)
    #---------------------------------------------
    if clean_dir
      @clean_dir_list = clean_dir.split(',')
      @clean_dir_list.each{|clean|
        if !(clean =~ /^\//)
          clean = File.expand_path(clean)
        end
        clean.sub!(/^#{@topdir}/, '')
        clean.sub!(/^#{@userdir}/, "/#{File.basename(@userdir)}")
	clean += '%'
        recorded_dir_list = 
          Directory.find(:all, :conditions=>["path like ?", clean])
        if recorded_dir_list.size < 1
          warn "  Warning: clean-dir \"#{clean}\" is not found.\n"
        end
        recorded_dir_list.each{ |dir|
          normal_msg "DB Records under \"#{dir.path}\" are deleted.\n"
          dir.destroy
        }
      }
    end

    #---------------------------------------------
    #  T[`pX̊Jnn_gbvfBNgȊOɂ
    #---------------------------------------------
    if search_dir
      @search_actual_dir = "#{search_dir}"
      if !(@search_actual_dir =~ /^\//)
        @search_actual_dir = File.expand_path(@search_actual_dir)
      end
      [@topdir, @userdir].each{|d|
        break if !(@search_actual_dir)
        [@search_actual_dir].each{|s|
          if test(?-, d, s)
            warn "  Warning: search_dir \"#{search_dir}\" is too upper directory. --search-dir option is ignored.\n"
            @search_actual_dir = false
            break
          end
        }
      }
    end

    if @search_actual_dir
      raise IOError, "  Error: \"#{@search_actual_dir}\" is not found.\n" if !(File.exist?(@search_actual_dir))
      raise IOError, "  Error: \"#{@search_actual_dir}\" is not a directory.\n" if !(File.directory?(@search_actual_dir))

      if @search_actual_dir =~ /^#{@topdir}/ then
        sdir = @search_actual_dir.sub(/^#{@topdir}/, '')
      elsif @search_actual_dir =~ /^#{@userdir}/ then
        sdir = @search_actual_dir.sub(/^#{@userdir}/, "/#{File.basename(@userdir)}")
      else
        raise ArgumentError, " Error: search_dir \"#{search_dir}\" is not included in #{@topdir} or #{@userdir}.\n"
      end

      @search_dir = sdir
      @searth_top_dir = File.dirname(sdir)
      normal_msg "\"#{@search_actual_dir}\" is set as top search directory.\n"
    else
      @search_actual_dir = false
      @search_dir = false
      @searth_top_dir = ROOTDIR_NAME
    end
  end

  #
  #̃\bhĂяo, G[ȊÕbZ[Wo͂sȂ悤ɂȂ.
  # false ^, ʏ̏o̓[hɖ߂.
  #
  def quiet(flag=true)
    if flag
      @verbose_level = 0
    else
      @verbose_level = 1
    end
  end

  #
  #̃\bhĂяo, i󋵂璷ɃbZ[Wo͂.
  # false ^, ʏ̏o̓[hɖ߂.
  #
  def verbose(flag=true)
    if flag
      @verbose_level = 2
    else
      @verbose_level = 1
    end
  end

  #
  #dirname ɗ^fBNgȉ̃fBNg YAML, SIGEN t@C
  #͂, f[^x[Xɓo^.
  #parent_id ɂ dirname ̃f[^x[X ID ^.
  #dirname ܂ parent_id ȗꍇɂ, new \bh
  #^ topdir fBNgƂ.
  #
  def register(dirname=nil, parent_id=nil, returned_dir=nil)

    #---------------------------------------------
    #  initialize  search_dir w肳Ăꍇ
    #---------------------------------------------
    if @search_actual_dir && @search_dir
      parent_dir_list = Directory.find_by_sql(["select * from directories where path like ? order by path",
                                                File.dirname(@search_dir)])
      if parent_dir_list.size > 0
        dirname = File.dirname(@search_dir)
        @search_dir = false
        parent_id = parent_dir_list[0].id
        Dir.chdir(File.dirname(@search_actual_dir))
      else
        raise ArgumentError, " Error: parent dir of search_dir \"#{@search_dir}\" has not been registered.\n"
      end
    end


    #---------------------------------------------
    #  gbvfBNg, [UfBNg, ȊO̔Ƌ
    #---------------------------------------------
    if !(parent_id) || !(dirname)
      recorded_dir_list = 
        Directory.find_by_sql(["select * from directories where path like ? order by path",
                                ROOTPATH_NAME])
      if recorded_dir_list.size > 0
        verbose_msg "Skip \"... #{ROOTPATH_NAME}\" on record.\n"
        parent_id = recorded_dir_list[0].id
      else
        topdir = Directory.new
        topdir.name = ROOTDIR_NAME
        topdir.path = ROOTPATH_NAME
        topdir.save
        normal_msg "."
        verbose_msg "Insert info from \"... /\".\n"
        parent_id = topdir.id
      end
      begin
        Dir.chdir(@topdir)
      rescue
        verbose_msg "Skip \"... #{ROOTPATH_NAME}\" for \"#{$!.to_s}\"\n"
        return false
      end
      dirname = ROOTDIR_NAME

      #---------------------------------------------
      #  [UfBNgֈړ
      #    (@search_actual_dir w肳ꍇ͈ړȂ).
      #---------------------------------------------
      if !(@search_actual_dir)
        toppwd = Dir.pwd
        Dir.chdir(File.dirname(@userdir))
        register(ROOTDIR_NAME + File.basename(@userdir), parent_id, toppwd)
      end

    elsif dirname == ROOTDIR_NAME + File.basename(@userdir)
      recorded_dir_list = 
        Directory.find_by_sql(["select * from directories where path like ? order by path",
                                dirname])
      if recorded_dir_list.size > 0
        verbose_msg "Skip \"... #{dirname}\" on record.\n"
        parent_id = recorded_dir_list[0].id
      else
        usrdir = Directory.new
        usrdir.name = File.basename(dirname)
        usrdir.path = dirname
        usrdir.parent_id = parent_id
        usrdir.save
        normal_msg "."
        verbose_msg "Insert info from \"... #{dirname}\".\n"
        parent_id = usrdir.id
      end
      if !(@search_actual_dir)
        begin
          Dir.chdir(File.basename(@userdir))
        rescue
          verbose_msg "Skip \"... #{dirname}\" for \"#{$!.to_s}\"\n"
          return false
        end
      end
    else
      if !(@search_actual_dir)
        begin
          Dir.chdir(File.basename(dirname))
        rescue
          verbose_msg "Skip \"... #{dirname}\" for \"#{$!.to_s}\"\n"
          return false
        end
      end
    end

    #---------------------------------------------
    #  o^
    #---------------------------------------------

    if !(@search_actual_dir)
      filelist = Dir.glob("*")
    elsif
      filelist = [File.basename(@search_actual_dir)]
      @search_actual_dir = false
    end

    filelist.each{ |file|
      if !File.readable?(file) || \
        File.directory?(file) && !(/mswin/ =~ Config::CONFIG["arch"]) && !File.executable?(file)
        verbose_msg "Skip \"... #{File.join(dirname, file)}\" for \"Permission denied\"\n"
        next
      end

      ignored_flag = false
      if GFDNAVI_IGNORED_DIRS
        GFDNAVI_IGNORED_DIRS.each{|igdir|
          if igdir.class.to_s == "String"
            ignored_flag = true if igdir == file
          elsif igdir.class.to_s == "Regexp"
            ignored_flag = true if igdir =~ file
          end
        }
      end
      next if ignored_flag
      if File.directory?(file)
        dir_path = File.join(dirname, file)
        recorded_dir_list = 
          Directory.find_by_sql(["select * from directories where path like ? order by path",
                                  dir_path])
        if recorded_dir_list.size > 0
          verbose_msg "Skip \"... #{dir_path}\" on record.\n"
          dir_id = recorded_dir_list[0].id
        else
          dir = Directory.new
          dir.name = file
          dir.path = dir_path
          dir.parent_id = parent_id
          dir.save
          normal_msg "."
          verbose_msg "Insert info from \"... #{dir.path}\".\n"
          dir_id = dir.id
        end
        meta_files_register(file, dir_id, dirname)
        # recursive search
        register(File.join(dirname, file), dir_id, Dir.pwd)
      elsif NumRu::GPhys::IO.file2type(file)
        gphys_reg = GPhysRegister.new(file, parent_id, dirname)
        next if gphys_reg.error
        gphys_reg.quiet if @verbose_level < 1
        gphys_reg.verbose if @verbose_level > 1
        dir_id = gphys_reg.register
        gphys_reg.close
        meta_files_register(file, dir_id, dirname)
      elsif file =~ NON_GPHYSIZABLE_SUFFIXES # 摜t@C̓o^.

        #---------------------------------------------
        #  Directory ̓o^
        #---------------------------------------------
        nong_directory_name = file.sub(NON_GPHYSIZABLE_SUFFIXES, '')
        nong_directory_path = File.join(dirname, nong_directory_name)

        recorded_dir = 
          Directory.find(:first, :conditions=>["path = ?",nong_directory_path])
        if recorded_dir
          verbose_msg "Skip \"... #{nong_directory_path}\" on record.\n"
          nongdir_id = recorded_dir.id
        else
          nongdir = Directory.new
          nongdir.name = nong_directory_name
          nongdir.path = nong_directory_path
          nongdir.parent_id = parent_id
          nongdir.save
          normal_msg "."
          verbose_msg "Insert info from \"... #{nong_directory_path}\".\n"
          nongdir_id = nongdir.id
        end

        #---------------------------------------------
        #  Variable ̓o^
        #---------------------------------------------

        recorded_var = 
          Variable.find(:first, :conditions=>["path = ?",
                                              File.join(dirname, file)])
        if recorded_var
          verbose_msg "Skip \"... #{File.join(dirname, file)}\" on record.\n"
          nongvar = recorded_var.id
        else
          nongvar = Variable.new
          nongvar.path = File.join(dirname, file)
          nongvar.name = file
          nongvar.directory_id = nongdir_id
          nongvar.gphysizable = false
          nongvar.save
          normal_msg "."
          verbose_msg "Insert info from \"... #{File.join(dirname, file)}\".\n"
          nongvar_id = nongvar.id
        end

        #---------------------------------------------
        #  KeywordAttribute ̓o^
        #---------------------------------------------

        meta_files_register(file, nongvar_id, dirname)
      end
    }

    #---------------------------------------------
    #  efBNgɕԂ or I
    #---------------------------------------------

    if returned_dir
      Dir.chdir(returned_dir)
    else
      normal_msg "\nDirectory tree search has been completed.\n" if dirname == @searth_top_dir
    end
  end

  #
  #^f[^i[t@Cǂݍ, fBNg
  #̂߂f[^Ƃēo^. , ȉ̎ނ̃t@C
  #ǂݍ. *file* ɂ̓^f[^t@C, *entity_id* ɂ
  #񂪕tfBNg (= fBNg, f[^t@C)
  # ID , *dirname* ɂ̓f[^x[X̃fBNg 
  #(t@CufBNgł͂ȂƂɒ) 
  #^.
  #
  def meta_files_register(file, entity_id, dirname)
    yaml_register(file+".yml", entity_id, dirname)
    GFDNAVI_LANGUAGES.each{ |lang|
      yaml_register(file+".yml."+lang, entity_id, dirname, lang)
    }
    sigen_register(file+".SIGEN", entity_id, dirname)
    GFDNAVI_LANGUAGES.each{ |lang|
      sigen_register(file+".SIGEN."+lang, entity_id, dirname, lang)
    }
  end

  #
  #yamlfile  YAML t@C^. entity_id ɂ YAML 
  #wfBNgt@C^. yamlfile 
  #Ȃꍇ͉Ȃł̃\bhI, e\bhɖ߂.
  #dirname ɂ̓f[^x[X̃fBNg
  #(t@CufBNgł͂ȂƂɒ) 
  #^.
  #
  def yaml_register(yamlfile, entity_id, dirname, lang=nil)
    return unless File.file?(yamlfile)
    new_registered_keyatt_ids = []
    path = File.join(dirname, yamlfile)
    begin
      yaml_info = YAML.load_file(yamlfile)
    rescue
      normal_msg "Skip \"... #{path}\" for \"#{$!.to_s}\"\n"
      return false
    end
    yaml_gphysvar_info = yaml_info.delete(YAML_GPHYSVAR_KEYWORD)
    new_registered_keyatt_ids <<
      RegisterSharedTools::keyword_attribute_register(yaml_info, entity_id, nil, lang)

    yaml_gphysvar_info.each{|varname, info|
      if varname =~ /(.*)[#{DIR_VAR_DELIMITER}\/](\w+)( +in +(\S+))?$/
        yaml_gphys_filename = $1
        yaml_gphysvar_name = $2
        yaml_gphys_alias = $4    # nil if ' as *' is not given (horinout)
        yaml_gphysdir_path = 
                   File.join(path.sub(/\.yml$/, '').sub(/.*\//, ''),
                        yaml_gphys_filename)
        yaml_gphys_files = Dir.glob(yaml_gphysdir_path).sort
        if yaml_gphys_files.size < 1
          raise IOError, "  Error: \"#{yaml_gphys_filename}\" in \"#{path}\" is not matched to actual files.\n"
        end

        if yaml_gphys_alias
          entity_id = nil   # directoryύX悤eID͕sƂ(horinout)
          yaml_gphysdir_path_in_db = 
                 File.expand_path(  \
                       File.join(path.sub(/\.yml$/, '').sub(/.*\//, ''), \
                                 yaml_gphys_alias) , '/'    \
                 ).sub(/^#{File.expand_path(".","\/")}/,'')
        else
          yaml_gphysdir_path_in_db = yaml_gphysdir_path
        end

        yaml_gphysdir = 
          GPhysRegister.new(yaml_gphys_files,
                            entity_id,
                            dirname,
                            yaml_gphysdir_path_in_db,
                            [yaml_gphysvar_name])

        next if yaml_gphysdir.error
        yaml_gphysdir.quiet if @verbose_level < 1
        yaml_gphysdir.verbose if @verbose_level > 1
        yaml_gphysdir_id = yaml_gphysdir.register
        yaml_gphysdir.close

        info.each{ |info_key, info_val|
          if info_val.class.to_s == "Hash" && info_key == yaml_gphysvar_name 
            yaml_gphysvar_list = 
              Variable.find_by_sql(["select * from variables where path like ? order by path",
                                     File.join(dirname, yaml_gphysdir_path_in_db) + DIR_VAR_DELIMITER + yaml_gphysvar_name])
            yaml_gphysvar_list.each{|yaml_gphysvar|
              new_registered_keyatt_ids <<
                RegisterSharedTools::keyword_attribute_register(info_val, nil, yaml_gphysvar.id, lang)
            }
          else
            new_registered_keyatt_ids <<
              RegisterSharedTools::keyword_attribute_register({info_key => info_val}, yaml_gphysdir_id, nil, lang)
          end
          }

      elsif path.sub(/\.yml$/, '') =~ NON_GPHYSIZABLE_SUFFIXES
        yaml_nongphysvar_path = path.sub(/\.yml$/, '')
        yaml_nongphysvar_list =
          Variable.find_by_sql(["select * from variables where path like ? order by path",
                                 yaml_nongphysvar_path])
        yaml_nongphysvar_list.each{|yaml_nongphysvar|
          new_registered_keyatt_ids <<
            RegisterSharedTools::keyword_attribute_register(info, nil, yaml_nongphysvar.id, lang)
        }
      else
        yaml_gphysvar_path = path.sub(/\.yml$/, '') +
          DIR_VAR_DELIMITER + varname
        yaml_gphysvar_list =
          Variable.find_by_sql(["select * from variables where path like ? order by path",
                                 yaml_gphysvar_path])
        yaml_gphysvar_list.each{|yaml_gphysvar|
          new_registered_keyatt_ids <<
            RegisterSharedTools::keyword_attribute_register(info, nil, yaml_gphysvar.id, lang)
        }
      end
    } if yaml_gphysvar_info
    if new_registered_keyatt_ids.flatten.size > 0
      normal_msg "."
      verbose_msg "Insert info from \"... #{path}\".\n"
    else
      verbose_msg "Skip \"... #{path}\" on record.\n"
    end
  end


  #
  #sigenfile  SIGEN t@C^. entity_id ɂ SIGEN 
  #wfBNgt@C^. sigenfile 
  #Ȃꍇ͉Ȃł̃\bhI, e\bhɖ߂.
  #dirname ɂ̓f[^x[X̃fBNg
  #(t@CufBNgł͂ȂƂɒ) 
  #^.
  #
  def sigen_register(sigenfile, entity_id, dirname, lang=nil)
    return unless File.file?(sigenfile)
    new_registered_keyatt_ids = []
    path = File.join(dirname, sigenfile)
    sigen = Sigen.new(sigenfile)
    sigen_info = sigen.parse
    new_registered_keyatt_ids <<
      RegisterSharedTools::keyword_attribute_register(sigen_info, entity_id, nil, lang)
    if new_registered_keyatt_ids.flatten.size > 0
      normal_msg "."
      verbose_msg "Insert info from \"... #{path}\".\n"
    else
      verbose_msg "Skip \"... #{path}\" on record.\n"
    end
  end

  def verbose_msg(msg)
    return if @verbose_level < 2
    print "  #{self.class}: #{msg}"
  end

  def normal_msg(msg)
    return if @verbose_level < 1
    print msg
    STDOUT.flush
  end

  #
  #GPhys[http://www.gfd-dennou.org/library/ruby/]
  #ňƂ\ȃf[^f[^x[XɊi[.
  #
  class GPhysRegister
    include NumRu

    attr_reader :error

    #
    #\bh. *file* ɂ String NX̃t@C, 
    #܂̓t@Ci[ Array NX^. Array NX
    #Ńt@C^ꍇ, X̃t@CʌɃf[^x[X
    #o^̂ł͂Ȃ, Array NXɋLqSẴt@C
    #zIȃt@Cf[^x[XւƓo^.
    #̍, zIȃt@ĆuĂіv *united_name*
    #ɗ^Kv (: "*.nc"). *united_name* ^Ȃ
    #ꍇ, Array NX̐擪̃t@Ĉ݂f[^x[Xւ
    #o^.
    #
    # *parent_id* ɂ *file* ̐efBNg
    #f[^x[X ID ^, *dirname* ɂ͐efBNg
    #f[^x[X̃pX^.
    #ɂ̓f[^x[X̃gbvfBNg^.
    #
    # *varnames* ͕ϐ Array NXɊi[ė^.
    #̈^ꂽꍇ, ̕ϐ݂̂f[^x[Xւ
    #i[.
    #
    def initialize(file, parent_id, dirname, united_name=nil, varnames=nil)
      @file_list = []
      @dirname = dirname
      @united_flag = false
      @varnames = nil
      @verbose_level = 1  # 0:quiet, 1: normal, 2: verbose
      if String === file
        @file_list << file
        @filename = @file_list[0]
      elsif Array === file
        if file.size < 1
          raise ArgumentError, " Error: line #{__LINE__} in #{__FILE__}, \"file\" argument is empty.\n"
        elsif String === united_name
          @file_list = file
          @filename = united_name
          @united_flag = true
        else
          @file_list << file[0]
          @filename = @file_list[0]
        end
      else
        raise ArgumentError, " Error: line #{__LINE__} in #{__FILE__}, \"file\" argument is " +
          " \"#{file.class.to_s} class. This class is invalid.\n"
      end
      if Array === varnames
        @varnames = []
        varnames.each{|v| @varnames << v.to_s}
      end

      @relative_path = File.join(@dirname, @filename)
      if parent_id
        @parent_id = parent_id
      else
        @parent_id = Directory.find_by_path(File.dirname(@relative_path)).id
      end
      @info = []
      @error = true
      @varlist_except_coordinates = []
      @file_obj_list = []
      @gphysdir_file_list = []
      @time_info = []
      @dir_longitude_info = []
      @dir_latitude_info = []

      @file_list.each{|file|
        rel_file_path = File.join(@dirname, file)
        begin
          @file_obj_list << GPhys::IO.file2file_class(file).open(file)
          @error = false
        rescue
          verbose_msg "Skip \"... #{rel_file_path}\" for \"#{$!.to_s}\"\n"
          next
        end
        varlist_except_coordinates_tmp = GPhys::IO.var_names_except_coordinates(@file_obj_list[-1])
        if @varlist_except_coordinates.size < 1
          @varlist_except_coordinates = varlist_except_coordinates_tmp
        else
          @varlist_except_coordinates.collect!{|checkv|
            if varlist_except_coordinates_tmp.include?(checkv)
              checkv
            else
              nil
            end
          }
          @varlist_except_coordinates.compact!
        end
        varlist = GPhys::IO.var_names(@file_obj_list[-1])
        gphysdir = GPhys::IO.open(@file_obj_list[-1], varlist[0])
        @gphysdir_file_list << gphysdir.data.file
        @time_info = self.parse_time(@file_obj_list[-1], varlist, @time_info)
        @dir_longitude_info = self.parse_longitude(@file_obj_list[-1], varlist, @dir_longitude_info)
        @dir_latitude_info = self.parse_latitude(@file_obj_list[-1], varlist, @dir_latitude_info)
      }

      if @varnames
        @varlist_except_coordinates.collect!{|checkv|
          if @varnames.include?(checkv)
            checkv
          else
            nil
          end
        }
        @varlist_except_coordinates.compact!
      end

    end

    #
    #I\bh. open Ă GPhys ϐ.
    #(GPhys ϐĴ 30 ܂łƂ炵.)
    #
    def close
      @file_obj_list.each{|o| o.close}
    end

    #
    #̃\bhĂяo, G[ȊÕbZ[Wo͂sȂ悤ɂȂ.
    # false ^, ʏ̏o̓[hɖ߂.
    #
    def quiet(flag=true)
      if flag
        @verbose_level = 0
      else
        @verbose_level = 1
      end
    end

    #
    #̃\bhĂяo, i󋵂璷ɃbZ[Wo͂.
    # false ^, ʏ̏o̓[hɖ߂.
    #
    def verbose(flag=true)
      if flag
        @verbose_level = 2
      else
        @verbose_level = 1
      end
    end

    #
    #new \bh *file* ɗ^f[^̏f[^x[X
    #i[.
    #
    #Ԃl͓o^fBNg ID ł.
    #
    def register
      dir_id = self.register_directory(@file_list.size==1)
      @varlist_except_coordinates.each{|varname|
        if @file_obj_list.size < 2
          file_obj = @file_obj_list[0]
        else
          file_obj = @file_obj_list
        end
        gphysvar = GPhysVariableRegister.new(file_obj, varname, dir_id, @relative_path)
        gphysvar.quiet if @verbose_level < 1
        gphysvar.verbose if @verbose_level > 1
        gphysvar.time_var(@time_info)
        gphysvar.longitude_var(@dir_longitude_info)
        gphysvar.latitude_var(@dir_latitude_info)
        varid = gphysvar.register
        register_actual_file(varid) if @file_list.size >=2
      }
      dir_id
    end

    def register_actual_file(varid)
      @file_list.each do |file|
        path = File.join(@dirname, file)
        recorded_af_list =
          ActualFile.find_by_sql(["select * from actual_files where path like ? order by path",
                                   path])
        if recorded_af_list.size > 0
          #verbose_msg "Skip \"... #{path}\" on record.\n"
          #return recorded_af_list[0].id
          next
        end

        af = ActualFile.new
        af.path = path
        af.variable_id = varid
        af.save
      end
    end

    #
    #new \bh *file* ɗ^f[^ɂ, eϐł͂Ȃ,
    #f[^Ŝ̏f[^x[XɊi[. Ԃl̓f[^x[X
    #̃fBNg ID ł.
    #
    def register_directory(plain_file)
      recorded_dir_list = 
        Directory.find_by_sql(["select * from directories where path like ? order by path",
                               @relative_path])
      if recorded_dir_list.size > 0
        verbose_msg "Skip \"... #{@relative_path}\" on record.\n"
        return recorded_dir_list[0].id
      end

      dir = Directory.new
      dir.name = File.basename(@filename)
      dir.path = @relative_path
      dir.parent_id = @parent_id
      dir.plain_file = plain_file
      dir.save

      dir_info = self.parse
      RegisterSharedTools::keyword_attribute_register(dir_info, dir.id)

      if @dir_longitude_info || @dir_latitude_info
        spatt = SpatialAttribute.new
        spatt.directory_id = dir.id
        if @dir_longitude_info
          spatt.longitude_lb = @dir_longitude_info[0]['longitude_lb']
          spatt.longitude_rt = @dir_longitude_info[0]['longitude_rt']
          spatt.lon_ax_name = @dir_longitude_info[0]['name']
        end
        if @dir_latitude_info
          spatt.latitude_lb = @dir_latitude_info[0]['latitude_lb']
          spatt.latitude_rt = @dir_latitude_info[0]['latitude_rt']
          spatt.lat_ax_name = @dir_latitude_info[0]['name']
        end
        spatt.save
      end

      normal_msg "."
      verbose_msg "Insert info from \"... #{dir.path}\".\n"
      dir.id
    end

    #
    #GPhys fBNg̓\bh. ͂e Hash NX
    #i[ĕԂ.
    #
    def parse
      @info = {}
      @gphysdir_file_list.each{|gphys_file|
        if gphys_file.methods.include?("each_att")
          gphys_file.each_att{|att|
            if !(@info[att.name])
              @info[att.name] = att.get
            end
          }
        end
      }
      @info
    end

    #
    #== WɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name      :: ϐ̖O (String NX)
    #starttime :: Jn       (DateTime NX)
    #enddtime  :: I       (DateTime NX)
    #
    #O *prev_info* ɂ͊ parse_time 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #starttime ͂菬Ȓl, endtime ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== W̔@
    #
    #ȉ̍WoxWƔF
    #
    #* units  NumRu::Units#reduce5 ɂĕϊ̂
    #  "86400 s @ 2002-01-01T00:00:00.00 +00:00" ̂悤ȃtH[}bg
    #  ƂȂĂ邱. NumRu::Units#reduce5 ɂĂ
    #  {NumRu::Units}[http://ruby.gfd-dennou.org/products/numru-units/doc/units.html]
    #  QƂ̂.
    #
    def parse_time(file_obj, varlist, prev_info=[])
      time_info = prev_info
      varlist.each{|varname|
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        next if gphysvar.length == 0
        gphysvar.att_names.each{|attname|
          if attname == 'units'
            units = Units.new(gphysvar.get_att(attname))
            begin
              units.reduce5.to_s
            rescue
              next
            end
            if units.reduce5.to_s =~ /^\((\d+ )?s @ (\d{4}\-\d{2}\-\d{2}T\d{2}:\d{2}:\d{2}\.\d{2} [\+\-]\d{2}:\d{2})\)$/
              info['name'] = varname
              unit_no_offset = $1 || 1
              unit_no_offset = unit_no_offset.to_f
              offset = DateTime.parse($2)
	      val = gphysvar.val
              start_val = val.min
              end_val = val.max
              info['starttime'] = offset + (start_val * (unit_no_offset / 86400.0))
              info['endtime'] = offset + (end_val * (unit_no_offset / 86400.0))
              time_info.collect!{|prev|
                if prev['name'] == info['name']
                  if prev['starttime'] < info['starttime']
                    info['starttime'] = prev['starttime']
                  end
                  if prev['endtime'] > info['endtime']
                    info['endtime'] = prev['endtime']
                  end
                  nil
                else
                  prev
                end
              }
              time_info.compact!
              time_info << info
            end
          end
        }
      }
      if time_info.size > 0
        time_info
      else
        nil
      end
    end

    #
    #== oxWɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #ox̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name         :: oxϐ̖O (String NX)
    #longitude_lb :: ox̐[     (Float NX)
    #longitude_rt :: ox̓[     (Float NX)
    #
    #O *prev_info* ɂ͊ parse_longitude 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #longitude_lb ͂菬Ȓl, longitude_rt ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== oxW̔@
    #
    #ȉ̍WoxWƔF
    #
    #* ϐ *lon* ܂ *longitude* ł邱 (啶͋ʂȂ)
    #* units  *degree_east* ܂ *degrees_east* ł邱.
    #
    #
    def parse_longitude(file_obj, varlist, prev_info=[])
      lon_info = prev_info
      varlist.each{|varname|
        flag = nil
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        next if gphysvar.length == 0
        if varname =~ /^lon(gitude)?$/i
          flag = true
        else
          gphysvar.att_names.each{|attname|
            if attname == 'units'
              if gphysvar.get_att(attname) =~ /^degrees?_east$/
                flag = true
              end
            end
          }
        end
        if flag
          info['name'] = varname
	  val = gphysvar.val
          info['longitude_lb'] =  val.min
          info['longitude_rt'] = val.max
          lon_info.collect!{|prev|
            if prev['name'] == info['name']
              if prev['longitude_lb'] < info['longitude_lb']
                info['longitude_lb'] = prev['longitude_lb']
              end
              if prev['longitude_rt'] > info['longitude_rt']
                info['longitude_rt'] = prev['longitude_rt']
              end
              nil
            else
              prev
            end
          }
          lon_info.compact!
          lon_info << info
        end
      }
      if lon_info.size > 0
        lon_info
      else
        nil
      end
    end

    #
    #== ܓxWɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #ܓx̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name        :: ܓxϐ̖O (String NX)
    #latitude_lb :: ܓx̓[     (Float NX)
    #latitude_rt :: ܓx̖k[     (Float NX)
    #
    #O *prev_info* ɂ͊ parse_latitude 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #latitude_lb ͂菬Ȓl, latitude_rt ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== ܓxW̔@
    #
    #ȉ̍WܓxWƔF
    #
    #* ϐ *lat* ܂ *latitude* ł邱 (啶͋ʂȂ)
    #* units  *degree_north* ܂ *degrees_north* ł邱.
    #
    #
    def parse_latitude(file_obj, varlist, prev_info=[])
      lat_info = prev_info
      varlist.each{|varname|
        flag = nil
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        next if gphysvar.length == 0
        if varname =~ /^lat(itude)?$/i
          flag = true
        else
          gphysvar.att_names.each{|attname|
            if attname == 'units'
              if gphysvar.get_att(attname) =~ /^degrees?_north$/
                flag = true
              end
            end
          }
        end
        if flag
          info['name'] = varname
	  val = gphysvar.val
	  info['latitude_lb'] =  val.min
	  info['latitude_rt'] = val.max
          lat_info.collect!{|prev|
            if prev['name'] == info['name']
              if prev['latitude_lb'] < info['latitude_lb']
                info['latitude_lb'] = prev['latitude_lb']
              end
              if prev['latitude_rt'] > info['latitude_rt']
                info['latitude_rt'] = prev['latitude_rt']
              end
              nil
            else
              prev
            end
          }
          lat_info.compact!
          lat_info << info
        end
      }
      if lat_info.size > 0
        lat_info
      else
        nil
      end
    end

    def verbose_msg(msg)
      return if @verbose_level < 2
      print "  #{self.class}: #{msg}"
    end

    def normal_msg(msg)
      return if @verbose_level < 1
      print msg
      STDOUT.flush
    end

  end

  class GPhysVariableRegister
    include NumRu
    #
    #\bh. *file* ɂ̓t@C, *varname* ɂ
    #ϐ^.
    #
    def initialize(file, varname, directory_id, directory_path)
      @varname = varname
      @gphysvar = GPhys::IO.open(file, @varname)
      @directory_id = directory_id
      @path = directory_path + DIR_VAR_DELIMITER + @varname
      @time_info = nil
      @dir_longitude_info = nil
      @dir_latitude_info = nil
      @verbose_level = 1  # 0:quiet, 1: normal, 2: verbose
    end

    #
    #̃\bhĂяo, G[ȊÕbZ[Wo͂sȂ悤ɂȂ.
    # false ^, ʏ̏o̓[hɖ߂.
    #
    def quiet(flag=true)
      if flag
        @verbose_level = 0
      else
        @verbose_level = 1
      end
    end

    #
    #̃\bhĂяo, i󋵂璷ɃbZ[Wo͂.
    # false ^, ʏ̏o̓[hɖ߂.
    #
    def verbose(flag=true)
      if flag
        @verbose_level = 2
      else
        @verbose_level = 1
      end
    end

    #
    #hash ɎԕϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐԕϐɈˑꍇ, ^ꂽ񂩂 starttime, endtime
    #Ƀf[^o^.
    #
    def time_var(info)
      @time_info = info
    end

    #
    #hash ɌoxϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐoxϐɈˑꍇ, ^ꂽ, spatial_attribute
    #e[uɃf[^o^.
    #
    def longitude_var(info)
      @dir_longitude_info = info
    end

    #
    #hash ɈܓxϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐܓxϐɈˑꍇ, ^ꂽ, spatial_attribute
    #e[uɃf[^o^.
    #
    def latitude_var(info)
      @dir_latitude_info = info
    end

    #
    #new \bh *file*, *varname* ɗ^ϐ̏f[^x[X
    #i[. Ԃlɕϐ̃f[^x[X ID Ԃ.
    #
    def register
      recorded_var_list = 
        Variable.find_by_sql(["select * from variables where path like ? order by path",
                               @path])
      if recorded_var_list.size > 0
        verbose_msg "Skip \"... #{@path}\" on record.\n"
        return recorded_var_list[0].id
      end

      var = Variable.new
      var.path = @path
      var.name = @varname
      var.directory_id = @directory_id
      var.gphysizable = true
      if @time_info
        @time_info.each{|info|
          if @gphysvar.axnames.include?(info['name'])
            var.starttime = info['starttime']
            var.endtime = info['endtime']
            break
          end
        }
      end
      var.save

      var_info = self.parse_attributes
      RegisterSharedTools::keyword_attribute_register(var_info, nil, var.id)

#      if @dir_longitude_info || @dir_latitude_info
#        spatt = SpatialAttribute.new
#        spatt.variable_id = var.id
#        if @dir_longitude_info
#          @dir_longitude_info.each{|info|
#            if @gphysvar.axnames.include?(info['name'])
#              spatt.longitude_lb = info['longitude_lb']
#              spatt.longitude_rt = info['longitude_rt']
#              spatt.lon_ax_name = info['name']
#            end
#          }
#        end
#        if @dir_latitude_info
#          @dir_latitude_info.each{|info|
#            if @gphysvar.axnames.include?(info['name'])
#              spatt.latitude_lb = info['latitude_lb']
#              spatt.latitude_rt = info['latitude_rt']
#              spatt.lat_ax_name = info['name']
#            end
#          }
#        end
#        spatt.save

      if @dir_longitude_info || @dir_latitude_info
        spatt = SpatialAttribute.new
        spatt.variable_id = var.id
        if @dir_longitude_info
          @dir_longitude_info.each{|info|
            spatt.longitude_lb = info['longitude_lb']
            spatt.longitude_rt = info['longitude_rt']
            spatt.lon_ax_name = info['name']
          }
        end
        if @dir_latitude_info
          @dir_latitude_info.each{|info|
            spatt.latitude_lb = info['latitude_lb']
            spatt.latitude_rt = info['latitude_rt']
            spatt.lat_ax_name = info['name']
          }
        end
        spatt.save


      end

      normal_msg "."
      verbose_msg "Insert info from \"... #{@path}\".\n"

      var.id
    end

    #
    #GPhys ϐ̑Ɋւ̓\bh. ͂e Hash NX
    #i[ĕԂ.
    #
    def parse_attributes
      @info = {}
      @gphysvar.att_names.each{|attname|
        @info[attname] = @gphysvar.get_att(attname)
      }
      @info
    end

    def verbose_msg(msg)
      return if @verbose_level < 2
      print "  #{self.class}: #{msg}"
    end

    def normal_msg(msg)
      return if @verbose_level < 1
      print msg
      STDOUT.flush
    end

  end

end


#
#fBNgf[^x[X}l[W
#mksigen[http://www.gfd-dennou.org/arch/cc-env/mksigen/SIGEN.htm]
# SIGEN t@C̉́E쐬pNX
#
class Sigen

  #
  #\bh. *file* ɂ̓t@C^.
  #
  #͂̏ꍇ͑ *rw* ^Kv͂Ȃ. *file* ݂Ȃ
  #ꍇ, G[Ԃ.
  #
  #쐬̏ꍇ͑ *rw*  "w"  "w+" ^.
  #
  def initialize(file, rw="r")
    @filename = File.expand_path(file)
    @io = open(@filename, rw)
    @rw = rw
    @info = {}
  end

  #
  #SIGEN t@C̓\bh. ͂e Hash ƂĕԂ.
  #Hash ̃L[ɃtB[h, lɂ̓tB[h{fBi[
  #. tB[h͏ɕϊ. tB[h{fB
  #̉s╡̋󔒕͑Sċ󔒈ꕶɕϊ.
  #tB[h, tB[h{fB̕R[h UTF-8 
  #ϊ.
  #
  def parse
    if @rw == "w"
      raise IOError, "  Error: #{@filename} was opened writable mode.\n"
    end
    name = ""
    @io.each{ |line|
      line.chomp!
      line = Kconv.kconv(line, GFDNAVI_CHARSET)
      break if line =~ /^\s*$/u
      if ! (line =~ /^\s/u)
        if ! (line =~ /^(\w+):\s*(.*)/u)
          warn "  Warning: broken header \"#{line}\", " +
            "line #{@io.lineno} in \"#{@filename}\"\n"
          name = ""
          next
        end
        name = $1.tr('A-Z', 'a-z')
        val = $2.gsub(/[\t\r]/u, ' ')
        if @info[name] then
          @info[name] << ' ' + val
        else
          @info[name] = val
        end
      else
        next if name.empty?
        @info[name] << line.gsub(/[\t\r]/u, ' ').gsub(/^\s*/u, ' ')
      end
    }
    @info
  end

  # SIGEN t@C쐬p\bh܂ĂȂ.
  def generate(sigen_hash)
    
  end
end



#
# ȉ͂̃t@CsvOƂēۂ̓.
#
if $0 == __FILE__
  opt = OptionParser.new
  OPTS = {}
  ARGV.options{|opt|
    opt.on( '-T=DIR', '--topdir=DIR',
            "Set data top directory
                                     (default: #{GFDNAVI_DATA_PATH})"
            ){|v| OPTS[:topdir] = v.gsub(/^=/, '')}

    opt.on( '-U=DIR', '--userdir=DIR',
            "Set user top directory
                                     (default: #{GFDNAVI_USER_PATH})"
            ){|v| OPTS[:userdir] = v.gsub(/^=/, '')}

    opt.on(         '--clean-dir=DIR',
            "Clean this directory once before register
                                     (example: /home/user/gfdnavi/public/data/analysis,/simulation)"
            ){|v| OPTS[:clean_dir] = v.gsub(/^=/, '')}

    opt.on(         '--search-dir=DIR',
            "Search directory
                                     (default: #{GFDNAVI_DATA_PATH})"
            ){|v| OPTS[:search_dir] = v.gsub(/^=/, '')}

    opt.on( '-P', '--production',
            "Use database \"#{CONFIG_DATABASE['production']['database']}\""
            ){|v| OPTS[:database] = 'production'}

    opt.on(       '--test',
            "Use database \"#{CONFIG_DATABASE['test']['database']}\"
                                     (default: \"#{CONFIG_DATABASE[DEFAULT_DATABASE]['database']}\")"
            ){|v| OPTS[:database] = 'test'}

    opt.on( '-q', '--quiet',
            "Quiet mode"
            ){|v| OPTS[:quiet] = v}

    opt.on( '-v', '--verbose',
            "Verbose mode"
            ){|v| OPTS[:verbose] = v}

    opt.on_tail('-h', '-H', '--help', 
                "This help message is output"
                ){|v| OPTS[:help] = v}
    opt.parse!
  }

  if OPTS[:help]
    print <<-"EOF"

  #{File.basename($0.to_s)}:

    USAGE: #{File.basename($0.to_s)} [options]

    OPTION: \n#{opt.to_a[1..-1].join("")}
    EOF
    exit
  end

  regdirtree = 
    RegisterDirTree.new(OPTS[:topdir], OPTS[:userdir], 
                        OPTS[:database], OPTS[:quiet], OPTS[:verbose], 
                        OPTS[:clean_dir], OPTS[:search_dir])
  regdirtree.register
  if !(OPTS[:clean_dir]) && !(OPTS[:search_dir])
    print "  Notice: for analysis of variable relation, parse again. (tentatively).\n"
    regdirtree.register
  end

end
