desc "Setup all"
task :setup => :environment do
  if Variable.table_exists? && Variable.count > 0
    print "This will clear all the database tables.\n"
    print "If you want just to update database tables, use \"rake update\" instead.\n"
    print "Are you sure to continue? (no/yes, default:no) "
    res = $stdin.gets.chomp!
    unless res == "yes"
      abort "execution was canceled"
    end
  end
  Rake::Task["setup:tables"].invoke
  Rake::Task["setup:tree"].invoke
  Rake::Task["setup:functions"].invoke
end

namespace :setup do
  desc "Setup database tables"
  task :tables => :environment do
    command = 'ruby "' + File.join(File.expand_path(RAILS_ROOT), 'db', 'dump_db.rb') + '"'
    unless system( command )
      raise "failed to execute '#{command}'"
    end
    pw = nil
    if User.table_exists? && (root = User.find_by_login('root'))
      print "Do you want to keep root's password? (yes/no, default:yes): "
      res = $stdin.gets.chomp!
      unless res == "no"
        pw = root.password
      end
    end

    pw2 = "temporary_password"
    if !pw
      if /mswin(?!ce)|mingw|bccwin/ =~ RUBY_PLATFORM
      else
        system("stty -echo > /dev/null 2>&1") #for unix-like systems
      end
      while true
        while true
          print "Input password for root (number of characters must be 5 to 40): "
          pw2 = $stdin.gets.chomp!
          if pw2.length < 5
            print "The password which you input is too short.\n"
          elsif pw2.length > 40
            pinnt "The password which you input is too long.\n"
          else
            break
          end
        end
        print "\nRepeat password: "
        tmp = $stdin.gets.chomp!
        break if tmp == pw2
        print "\nNot correct!\n"
      end
      if /mswin(?!ce)|mingw|bccwin/ =~ RUBY_PLATFORM
      else
        system("stty echo > /dev/null 2>&1") #for unix-like systems
      end
      print("\n")
    end

    connection = ActiveRecord::Base.connection
    connection.drop_table("schema_info") if connection.tables.include?("schema_info")
    Rake::Task["db:migrate"].invoke
    root = User.new( :login => "root",
		     :full_name => "default super user",
		     :email_address => GFDNAVI_ADMIN_EMAIL,
		     :affiliation => "GFDNAVI",
                     :password => pw2,
                     :password_confirmation => pw2 )
    root.super_user = true
    root.save!
    if pw
      root.password = pw
      root.save!
    end

    group = Group.new( :name => "group_missing",
                       :user_id => root.id )
    group.save!
    gm = GroupMember.new( :user_id => root.id,
                          :group_id => group.id )
    gm.save!
  end

  desc "Set root password"
  task :root_password => :environment do
    if root = User.find_by_login("root")
      if /mswin(?!ce)|mingw|bccwin/ =~ RUBY_PLATFORM
      else
        system("stty -echo > /dev/null 2>&1") #for unix-like systems
      end
      while true
        while true
          print "Input password for root (number of characters must be 5 to 40): "
          pw2 = $stdin.gets.chomp!
          if pw2.length < 5
            print "The password which you input is too short.\n"
          elsif pw2.length > 40
            pinnt "The password which you input is too long.\n"
          else
            break
          end
        end
        print "\nRepeat password: "
        tmp = $stdin.gets.chomp!
        break if tmp == pw2
        print "\nNot correct!\n"
      end
      if /mswin(?!ce)|mingw|bccwin/ =~ RUBY_PLATFORM
      else
        system("stty echo > /dev/null 2>&1") #for unix-like systems
      end
      print("\n")
      if root.change_password(pw2,tmp)
        print "changing was succeded\n"
      else
        print "changing was not succeded\n"
      end
    end
  end

  desc "Setup tree of directories and variables"
  task :tree => :environment do
    [ DrawParameter, KeywordAttribute, SpatialAttribute,
      VariableRelation, ActualFile,
      Variable, Directory].each{|model|
	model.delete_all if model.table_exists?
    }
    command = 'ruby "' + File.join(File.expand_path(RAILS_ROOT), 'db', 'register_directory_tree.rb') + '"'
    unless system( command )
      raise "failed to execute '#{command}'"
    end
  end

  desc "Setup functions database"
  task :functions => :environment do
    fixture = File.join(File.expand_path(RAILS_ROOT), 'db', 'function_defaults.yml')
    functions = YAML::load( File.read(fixture) )
    FunctionArgument.delete_all if FunctionArgument.table_exists?
    Type.delete_all if Type.table_exists?
    Type.new(:name => "int").save!
    Type.new(:name => "float").save!
    Type.new(:name => "string").save!
    FunctionOutput.delete_all if FunctionOutput.table_exists?
    Function.delete_all if Function.table_exists?
    functions.each{|k,v|
      arguments = v.delete(:arguments)
      outputs = v.delete(:outputs)
      func = Function.new(v)
      func.save!
      func.function_outputs.create(outputs)
      if arguments
        arguments.each{|arg|
          type = arg.delete(:type)
          fa = FunctionArgument.new(arg)
          fa.type = Type.find_by_name(type)
          fa.function = func
          fa.save!
        }
      end
    }
  end

end
